// From http://www.redblobgames.com/grids/hexagons/
// Copyright 2015 Red Blob Games <redblobgames@gmail.com>
// License: Apache v2.0 <http://www.apache.org/licenses/LICENSE-2.0.html>
//
// These are the hex grid algorithms expressed in Haxe syntax, to be
// converted to an AST for translation to another language.

import Ast.*;
using Ast;

class Codegen {
    static public function pointClass() {
        return classdef
            ({
                var x:Float;
                var y:Float;
            });
    }


    static public function orientationClass() {
        return classdef
            ({
                var f0:Float;
                var f1:Float;
                var f2:Float;
                var f3:Float;
                var b0:Float;
                var b1:Float;
                var b2:Float;
                var b3:Float;
                var start_angle:Float;
            });
    }


    static public function layoutClass() {
        return classdef
            ({
                var orientation:Orientation;
                var size:Point;
                var origin:Point;

                var layout_pointy:Orientation =
                new Orientation(sqrt(3.0), sqrt(3.0)/2.0, 0.0, 3.0/2.0,
                                sqrt(3.0)/3.0, -1.0/3.0, 0.0, 2.0/3.0,
                                0.5);

                var layout_flat:Orientation =
                new Orientation(3.0/2.0, 0.0, sqrt(3.0)/2.0, sqrt(3.0),
                                2.0/3.0, 0.0, -1.0/3.0, sqrt(3.0)/3.0,
                                0.0);

                function hex_to_pixel(layout:Layout, h:Hex):Point {
                    var M:Orientation = layout.orientation;
                    var size:Point = layout.size;
                    var origin:Point = layout.origin;
                    var x:Float = (M.f0 * Float(h.q) + M.f1 * Float(h.r)) * size.x;
                    var y:Float = (M.f2 * Float(h.q) + M.f3 * Float(h.r)) * size.y;
                    return new Point(x + origin.x, y + origin.y);
                }

                function pixel_to_hex(layout:Layout, p:Point):FractionalHex {
                    var M:Orientation = layout.orientation;
                    var size:Point = layout.size;
                    var origin:Point = layout.origin;
                    var pt:Point = new Point((p.x - origin.x) / size.x, (p.y - origin.y) / size.y);
                    var q:Float = M.b0 * pt.x + M.b1 * pt.y;
                    var r:Float = M.b2 * pt.x + M.b3 * pt.y;
                    return new FractionalHex(q, r, -q-r);
                }

                function hex_corner_offset(layout:Layout, corner:Int):Point {
                    var M:Orientation = layout.orientation;
                    var size:Point = layout.size;
                    // By using -corner instead of +corner I can make the corners
                    // line up with the dirs in hex_neighbor
                    var angle:Float = 2.0 * PI * (M.start_angle - Float(corner)) / 6.0;
                    return new Point(size.x * cos(angle), size.y * sin(angle));
                }
                
                // TODO: is it feasible to reuse hex_to_pixel, with
                // fractional points? It ends up being one third of 
                // the distance to a diagonal neighbor.
                function polygon_corners(layout:Layout, h:Hex):Array<Point> {
                    var corners:Array<Point> = new Array<Point>();
                    var center:Point = hex_to_pixel(layout, h);
                    for (i in 0...6) {
                        var offset:Point = hex_corner_offset(layout, i);
                        corners.push(new Point(center.x + offset.x, center.y + offset.y));
                    }
                    return corners;
                }
            });
    }

    
    static public function hexClass() {
        return classdef
            ({
                var q:Int;
                var r:Int;
                var s:Int;

                PRECONDITION(q + r + s != 0, "q + r + s must be 0");
                
                function hex_add(a:Hex, b:Hex):Hex {
                    return new Hex(a.q + b.q, a.r + b.r, a.s + b.s);
                }

                function hex_subtract(a:Hex, b:Hex):Hex {
                    return new Hex(a.q - b.q, a.r - b.r, a.s - b.s);
                }
                
                function hex_scale(a:Hex, k:Int):Hex {
                    return new Hex(a.q * k, a.r * k, a.s * k);
                }

                function hex_rotate_left(a:Hex):Hex {
                    return new Hex(-a.s, -a.q, -a.r);
                }
                
                function hex_rotate_right(a:Hex):Hex {
                    return new Hex(-a.r, -a.s, -a.q);
                }
                
                var hex_directions:Array<Hex> =
                    new Array<Hex>(
                                   new Hex(1, 0, -1), new Hex(1, -1, 0), new Hex(0, -1, 1),
                                   new Hex(-1, 0, 1), new Hex(-1, 1, 0), new Hex(0, 1, -1)
                                   );
                
                function hex_direction(direction:Int):Hex {
                    return hex_directions[direction];
                }

                function hex_neighbor(hex:Hex, direction:Int):Hex {
                    return hex_add(hex, hex_direction(direction));
                }

                var hex_diagonals:Array<Hex> =
                    new Array<Hex>(
                                   new Hex(2, -1, -1), new Hex(1, -2, 1), new Hex(-1, -1, 2),
                                   new Hex(-2, 1, 1), new Hex(-1, 2, -1), new Hex(1, 1, -2)
                                   );

                function hex_diagonal_neighbor(hex:Hex, direction:Int):Hex {
                    return hex_add(hex, hex_diagonals[direction]);
                }

                function hex_length(hex:Hex):Int {
                    return Int((abs(hex.q) + abs(hex.r) + abs(hex.s)) / 2);
                }
                
                function hex_distance(a:Hex, b:Hex):Int {
                    return hex_length(hex_subtract(a, b));
                }
            });
    }


    static public function fractionalHexClass() {
        return classdef
            ({
                var q:Float;
                var r:Float;
                var s:Float;

                PRECONDITION(round(q + r + s) != 0, "q + r + s must be 0");
                
                function hex_round(h:FractionalHex):Hex {
                    var qi:Int = Int(round(h.q));
                    var ri:Int = Int(round(h.r));
                    var si:Int = Int(round(h.s));

                    var q_diff:Float = abs(Float(qi) - h.q);
                    var r_diff:Float = abs(Float(ri) - h.r);
                    var s_diff:Float = abs(Float(si) - h.s);

                    if (q_diff > r_diff && q_diff > s_diff) {
                        qi = -ri-si;
                    } else if (r_diff > s_diff) {
                        ri = -qi-si;
                    } else {
                        si = -qi-ri;
                    }
                        
                    return new Hex(qi, ri, si);
                }

                function hex_lerp(a:FractionalHex, b:FractionalHex, t:Float):FractionalHex {
                    return new FractionalHex(a.q * (1.0 - t) + b.q * t,
                                             a.r * (1.0 - t) + b.r * t,
                                             a.s * (1.0 - t) + b.s * t);
                }

                function hex_linedraw(a:Hex, b:Hex):Array<Hex> {
                    var N:Int = hex_distance(a, b);
                    var a_nudge:FractionalHex = new FractionalHex(Float(a.q) + 1e-6, Float(a.r) + 1e-6, Float(a.s) - 2e-6);
                    var b_nudge:FractionalHex = new FractionalHex(Float(b.q) + 1e-6, Float(b.r) + 1e-6, Float(b.s) - 2e-6);
                    var results:Array<Hex> = new Array<Hex>();
                    var step:Float = 1.0 / Float(max(N, 1));
                    for (i in 0...N+1) {
                        results.push(hex_round(hex_lerp(a_nudge, b_nudge, step * Float(i))));
                    }
                    return results;
                }
            });
    }


    static public function offsetClass() {
        return classdef
            ({
                var col:Int;
                var row:Int;
                
                var EVEN:Int = 1;
                var ODD:Int = -1;
                
                function qoffset_from_cube(offset:Int, h:Hex):OffsetCoord {
                    var col:Int = h.q;
                    var row:Int = h.r + Int((h.q + offset * (h.q & 1)) / 2);
                    if (offset != EVEN && offset != ODD) { $error("offset must be EVEN (+1) or ODD (-1)"); }
                    return new OffsetCoord(col, row);
                }
                
                function qoffset_to_cube(offset:Int, h:OffsetCoord):Hex {
                    var q:Int = h.col;
                    var r:Int = h.row - Int((h.col + offset * (h.col & 1)) / 2);
                    var s:Int = -q-r;
                    if (offset != EVEN && offset != ODD) { $error("offset must be EVEN (+1) or ODD (-1)"); }
                    return new Hex(q, r, s);
                }

                function roffset_from_cube(offset:Int, h:Hex):OffsetCoord {
                    var col:Int = h.q + Int((h.r + offset * (h.r & 1)) / 2);
                    var row:Int = h.r;
                    if (offset != EVEN && offset != ODD) { $error("offset must be EVEN (+1) or ODD (-1)"); }
                    return new OffsetCoord(col, row);
                }
                
                function roffset_to_cube(offset:Int, h:OffsetCoord):Hex {
                    var q:Int = h.col - Int((h.row + offset * (h.row & 1)) / 2);
                    var r:Int = h.row;
                    var s:Int = -q-r;
                    if (offset != EVEN && offset != ODD) { $error("offset must be EVEN (+1) or ODD (-1)"); }
                    return new Hex(q, r, s);
                }
            });
    }

    
    static public function doubledClass() {
        return classdef
            ({
                var col:Int;
                var row:Int;
                
                function qdoubled_from_cube(h:Hex):DoubledCoord {
                    var col:Int = h.q;
                    var row:Int = 2 * h.r + h.q;
                    return new DoubledCoord(col, row);
                }
                
                function qdoubled_to_cube(h:DoubledCoord):Hex {
                    var q:Int = h.col;
                    var r:Int = Int((h.row - h.col) / 2);
                    var s:Int = -q-r;
                    return new Hex(q, r, s);
                }

                function rdoubled_from_cube(h:Hex):DoubledCoord {
                    var col:Int = 2 * h.q + h.r;
                    var row:Int = h.r;
                    return new DoubledCoord(col, row);
                }
                
                function rdoubled_to_cube(h:DoubledCoord):Hex {
                    var q:Int = Int((h.col - h.row) / 2);
                    var r:Int = h.row;
                    var s:Int = -q-r;
                    return new Hex(q, r, s);
                }
            });
    }

    
    static public function testsClass() {
        var testFramework = classdef
            ({
                function equal_hex(name:String, a:Hex, b:Hex):Void {
                    if (!(a.q == b.q && a.s == b.s && a.r == b.r)) { complain(name); } else {};
                }

                function equal_offsetcoord(name:String, a:OffsetCoord, b:OffsetCoord):Void {
                    if (!(a.col == b.col && a.row == b.row)) { complain(name); } else {}
                }

                function equal_doubledcoord(name:String, a:DoubledCoord, b:DoubledCoord):Void {
                    if (!(a.col == b.col && a.row == b.row)) { complain(name); } else {}
                }

                function equal_int(name:String, a:Int, b:Int):Void {
                    if (!(a == b)) { complain(name); } else {}
                }

                function equal_hex_array(name:String, a:Array<Hex>, b:Array<Hex>):Void {
                    equal_int(name, $array_length(a), $array_length(b));
                    for (i in 0...$array_length(a)) {
                        equal_hex(name, a[i], b[i]);
                    }
                }
            });
        
        var hexTests = classdef
            ({
                function test_hex_arithmetic():Void {
                    equal_hex("hex_add", new Hex(4, -10, 6), hex_add(new Hex(1, -3, 2), new Hex(3, -7, 4)));
                    equal_hex("hex_subtract", new Hex(-2, 4, -2), hex_subtract(new Hex(1, -3, 2), new Hex(3, -7, 4)));
                }

                function test_hex_direction():Void {
                    equal_hex("hex_direction", new Hex(0, -1, 1), hex_direction(2));
                }
             
                function test_hex_neighbor():Void {
                    equal_hex("hex_neighbor", new Hex(1, -3, 2), hex_neighbor(new Hex(1, -2, 1), 2));
                }

                function test_hex_diagonal():Void {
                    equal_hex("hex_diagonal", new Hex(-1, -1, 2), hex_diagonal_neighbor(new Hex(1, -2, 1), 3));
                }
                
                function test_hex_distance():Void {
                    equal_int("hex_distance", 7, hex_distance(new Hex(3, -7, 4), new Hex(0, 0, 0)));
                }

                function test_hex_rotate_right():Void {
                    equal_hex("hex_rotate_right", hex_rotate_right(new Hex(1, -3, 2)), new Hex(3, -2, -1));
                }
                
                function test_hex_rotate_left():Void {
                    equal_hex("hex_rotate_left", hex_rotate_left(new Hex(1, -3, 2)), new Hex(-2, -1, 3));
                }
                
                function test_hex_round():Void {
                    var a:FractionalHex = new FractionalHex(0.0, 0.0, 0.0);
                    var b:FractionalHex = new FractionalHex(1.0, -1.0, 0.0);
                    var c:FractionalHex = new FractionalHex(0.0, -1.0, 1.0);

                    // Test with a point on a hex
                    equal_hex("hex_round 1", new Hex(5, -10, 5), hex_round(hex_lerp(new FractionalHex(0.0, 0.0, 0.0), new FractionalHex(10.0, -20.0, 10.0), 0.5)));
                    // Test with a point between two hexes
                    equal_hex("hex_round 2", hex_round(a), hex_round(hex_lerp(a, b, 0.499)));
                    equal_hex("hex_round 3", hex_round(b), hex_round(hex_lerp(a, b, 0.501)));
                    // Test with a point between three hexes
                    equal_hex("hex_round 4", hex_round(a), hex_round(new FractionalHex(a.q*0.4+b.q*0.3+c.q*0.3,
                                                                                       a.r*0.4+b.r*0.3+c.r*0.3,
                                                                                       a.s*0.4+b.s*0.3+c.s*0.3)));
                    equal_hex("hex_round 5", hex_round(c), hex_round(new FractionalHex(a.q*0.3+b.q*0.3+c.q*0.4,
                                                                                       a.r*0.3+b.r*0.3+c.r*0.4,
                                                                                       a.s*0.3+b.s*0.3+c.s*0.4)));
                }

                function test_hex_linedraw():Void {
                    // NOTE: I've picked a line that doesn't have any
                    // tie breaking issues. Tie breaks can go either
                    // way, and no one answer is correct.
                    equal_hex_array("hex_linedraw",
                                     new Array<Hex>(new Hex(0, 0, 0), new Hex(0, -1, 1),
                                                     new Hex(0, -2, 2), new Hex(1, -3, 2),
                                                     new Hex(1, -4, 3), new Hex(1, -5, 4)),
                                     hex_linedraw(new Hex(0, 0, 0), new Hex(1, -5, 4)));
                }

                // NOTE: this tests the round trip but doesn't make sure the actual layout is correct
                function test_layout():Void {
                    var h:Hex = new Hex(3, 4, -7);
                    
                    var flat:Layout = new Layout(layout_flat, new Point(10.0, 15.0), new Point(35.0, 71.0));
                    equal_hex("layout", h, hex_round(pixel_to_hex(flat, hex_to_pixel(flat, h))));
                    
                    var pointy:Layout = new Layout(layout_pointy, new Point(10.0, 15.0), new Point(35.0, 71.0));
                    equal_hex("layout", h, hex_round(pixel_to_hex(pointy, hex_to_pixel(pointy, h))));
                }
            });

        var offsetTests =
            classdef
            ({
                function test_offset_roundtrip():Void {
                    var a:Hex = new Hex(3, 4, -7);
                    var b:OffsetCoord = new OffsetCoord(1, -3);
                    equal_hex("conversion_roundtrip even-q", a, qoffset_to_cube(EVEN, qoffset_from_cube(EVEN, a)));
                    equal_offsetcoord("conversion_roundtrip even-q", b, qoffset_from_cube(EVEN, qoffset_to_cube(EVEN, b)));
                    equal_hex("conversion_roundtrip odd-q", a, qoffset_to_cube(ODD, qoffset_from_cube(ODD, a)));
                    equal_offsetcoord("conversion_roundtrip odd-q", b, qoffset_from_cube(ODD, qoffset_to_cube(ODD, b)));
                    equal_hex("conversion_roundtrip even-r", a, roffset_to_cube(EVEN, roffset_from_cube(EVEN, a)));
                    equal_offsetcoord("conversion_roundtrip even-r", b, roffset_from_cube(EVEN, roffset_to_cube(EVEN, b)));
                    equal_hex("conversion_roundtrip odd-r", a, roffset_to_cube(ODD, roffset_from_cube(ODD, a)));
                    equal_offsetcoord("conversion_roundtrip odd-r", b, roffset_from_cube(ODD, roffset_to_cube(ODD, b)));
                }
                
                function test_offset_from_cube():Void {
                    equal_offsetcoord("offset_from_cube even-q", new OffsetCoord(1, 3), qoffset_from_cube(EVEN, new Hex(1, 2, -3)));
                    equal_offsetcoord("offset_from_cube odd-q", new OffsetCoord(1, 2), qoffset_from_cube(ODD, new Hex(1, 2, -3)));
                    // TODO: add more tests
                }

                function test_offset_to_cube():Void {
                    equal_hex("offset_to_cube even-", new Hex(1, 2, -3), qoffset_to_cube(EVEN, new OffsetCoord(1, 3)));
                    equal_hex("offset_to_cube odd-q", new Hex(1, 2, -3), qoffset_to_cube(ODD, new OffsetCoord(1, 2)));
                }
            });
        
        
        var doubledTests =
            classdef
            ({
                function test_doubled_roundtrip():Void {
                    var a:Hex = new Hex(3, 4, -7);
                    var b:DoubledCoord = new DoubledCoord(1, -3);
                    equal_hex("conversion_roundtrip doubled-q", a, qdoubled_to_cube(qdoubled_from_cube(a)));
                    equal_doubledcoord("conversion_roundtrip doubled-q", b, qdoubled_from_cube(qdoubled_to_cube(b)));
                    equal_hex("conversion_roundtrip doubled-r", a, rdoubled_to_cube(rdoubled_from_cube(a)));
                    equal_doubledcoord("conversion_roundtrip doubled-r", b, rdoubled_from_cube(rdoubled_to_cube(b)));
                }
                
                function test_doubled_from_cube():Void {
                    equal_doubledcoord("doubled_from_cube doubled-q", new DoubledCoord(1, 5), qdoubled_from_cube(new Hex(1, 2, -3)));
                    equal_doubledcoord("doubled_from_cube doubled-r", new DoubledCoord(4, 2), rdoubled_from_cube(new Hex(1, 2, -3)));
                }

                function test_doubled_to_cube():Void {
                    equal_hex("doubled_to_cube doubled-q", new Hex(1, 2, -3), qdoubled_to_cube(new DoubledCoord(1, 5)));
                    equal_hex("doubled_to_cube doubled-r", new Hex(1, 2, -3), rdoubled_to_cube(new DoubledCoord(4, 2)));
                }
            });
        
        
        // Add another function that calls all the other test functions
        var decls = testFramework.concat(hexTests).concat(offsetTests).concat(doubledTests);
        var testAllCode = [];
        for (decl in decls) {
            switch (decl) {
                case DMethod(name, _, [], _):
                    testAllCode.push(AFunctionCall(name, []));
                case _:
            }
        }
        decls.push(DMethod("test_all", TVoid, [], ABlock(testAllCode)));
        
        return decls;
    }
}
